#include	<stdio.h>
#include	<string.h>
#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	"lcd.h"
#include	"usart.h"
#include	"rpm.h"

//  :   ȣ ̺
#define	X	PB7
#define	Y	PB6
#define	MOTOR_PORT	PORTB
#define	MOTOR_DDR	DDRB
#define	CLAER	(MOTOR_PORT & 0x3F)

//   ũ 
#define	MOTOR_FOREWARD	(MOTOR_PORT = (MOTOR_PORT & CLAER) | 1<<X)
#define	MOTOR_BACKWARD	(MOTOR_PORT = (MOTOR_PORT & CLAER) | 1<<Y)
#define	MOTOR_STOP		(MOTOR_PORT = (MOTOR_PORT & CLAER) | 1<<X | 1<<Y)

// Kp, Ki, Kd ʱ갪 
#define INIT_Kp		2.0
#define INIT_Ki		1.0
#define	INIT_Kd		0.1

//    
#define	COMMAND_LEN	10
#define	GO		1
#define	STOP	0
#define	LOCAL_CONTROL	0
#define	PC_CONTROL		1

#define	MAX_RPM		150

// usart1 printf(), scanf() Լ  
static FILE usart1_stdio = FDEV_SETUP_STREAM(USART1_send, USART1_receive, _FDEV_SETUP_RW);

// kp, ki, kd   
double	kp = INIT_Kp, pc_kp = INIT_Kp;
double	ki = INIT_Ki, pc_ki = INIT_Ki;
double	kd = INIT_Kd, pc_kd = INIT_Kd;

// Ÿ rpm (local, pc, target) 
volatile double	local_target_rpm = 0.0;
double	pc_target_rpm = 40.0, target_rpm = 0.0;

// PID    
double		u;
double		prev_e, e=0., esum=0.;
double		dedt, ki_dt, dt;
volatile unsigned char	pid_req = 0;

//   
int		control = LOCAL_CONTROL;

//  LOCAL  AD ȯ  local Ÿ  
ISR(ADC_vect)	{	local_target_rpm = (long)MAX_RPM * ADC / 1024.0;	}

// ڰ  ˻,    ڿ  command[] о.
int	get_command(char command[])
{	int	flag;

	if( (flag = USART1_rx_check()) )
		scanf("%s", command);

	return flag;
}

//-------------- main() Լ ------------------------------
int	main(void)
{	char	lcd_string[2][MAX_LCD_STRING];
	char	command[COMMAND_LEN];
	int		flag = 0, run = STOP;

	LCD_init();					// LCD ʱȭ
	ICP1_input_capture_init();	// ICP1  Է ĸó ʱȭ
	TCNT1_PWM_10bit_init();		// Ÿ̸/ī1 10bit OC1A PWM ʱȭ
	USART_init(USART1, 12);		// USART1 Ʈ : 38400(UBRR=12) at 8MHz Ŭ

	ADCSRA |= 1<<ADEN | 1<<ADIE | 1<<ADSC | 6;	// LOCAL Ÿ rpm  AD ȯ 
	stdin = stdout = stderr = &usart1_stdio;	// printf(), scanf() usart1_stdio 
	sei();

	MOTOR_DDR |= 1<<X | 1<<Y;	//   Ʈ X, Y   
	MOTOR_BACKWARD;

	sprintf(lcd_string[0], "PID ctl ");
	LCD_str_write(0, 0, lcd_string[0]);

	esum = 0.;
	dt = SEC_PID_INTERVAL;

	pid_req = 1;		// Ÿ̸/ī1 ͷƮ  ƾ PID û
	while(1){
	  flag = get_command(command);
		
	  if(flag){
		if( strcmp(command, "GO")==0 )			run = GO;
		else if( strcmp(command, "STOP")==0 )	run = STOP;
		else if( strcmp(command, "PC")==0 )		control = PC_CONTROL;
		else if( strcmp(command, "LOCAL")==0 )	control = LOCAL_CONTROL;
		else if( strcmp(command, "PC_RPM")==0 )	scanf("%lf", &pc_target_rpm);	
		else if( strcmp(command, "KP")==0 ){
			scanf("%lf", &pc_kp);
			kp = (control==PC_CONTROL) ? pc_kp : kp; 
		}else if( strcmp(command, "KI")==0 ){
			scanf("%lf", &pc_ki);
			ki = (control==PC_CONTROL) ? pc_ki : ki; 
		}else if( strcmp(command, "KD")==0 ){
			scanf("%lf", &pc_kd);
			kd = (control==PC_CONTROL) ? pc_kd : kd; 
		}
	  }

	  if(pid_req==0){			// Ÿ̸/ī1 ͷƮ  ƾ PID ð  
	  	pid_req = 1;			//  û

	  	if(run==GO){	// 
			// LOCAL, PC  ο  target_rpm , AD ȯ 
			target_rpm = (control == PC_CONTROL) ? pc_target_rpm : local_target_rpm;
			ADCSRA |= (control == PC_CONTROL) ? 0 : (1<<ADIE | 1<<ADSC);	// LOCAL  AD ȯ 

			e = target_rpm - rpm;				// e : error
			esum = esum + e;					// esum : error 
			u = kp * e + ki * esum * dt + kd * (e - prev_e)/dt;		// u = pwm
			prev_e = e;

			OCR1A = (u >= (double)(OVF_TCOUNT-1)) ? (unsigned int)(OVF_TCOUNT-1) : (u <= 0.) ? 0 : (unsigned int)u;

			sprintf(lcd_string[0], "PWM:%-3d  ", (int)(u*100/(OVF_TCOUNT-1)));	// u : 0 ~ 100%  ȯ  
			sprintf(lcd_string[1], "Tgt:%-3d RPM:%-3d ", (int)target_rpm, (int)rpm);	// Ÿ rpm,  rpm 

			LCD_str_write(0, 8, lcd_string[0]);		// LCD 
			LCD_str_write(1, 0, lcd_string[1]);

			printf("%s %s\r\n", lcd_string[0], lcd_string[1]);	// ø 
	 	}else{		// Ͻ 
	  		OCR1A = 0;
			esum = 0;
	 	}
	  }
	}
}
